
#include "axllib"

define Adjoint(ObjA:Category,ObjB:Category, Left:ObjA->ObjB, Right:ObjB->ObjA):Category == with {
    >>: (A:ObjA,B:ObjB,  Left A ->       B ) -> (      A -> Right B );  -- 1/2 of the adjunction isomorphism
    <<: (A:ObjA,B:ObjB,       A -> Right B ) -> ( Left A ->       B );  -- 2/2 of the adjunction isomorphism
    
    unit:   (A:ObjA) -> (           A -> Right  Left A);    --   unit natural transformation
    counit: (B:ObjB) -> (Left Right B ->             B);    -- counit natural transformation
    
    left:  (X:ObjA,Y:ObjA,X->Y) -> (  Left X ->  Left Y );  --  Left functor action on morphisms
    right: (X:ObjB,Y:ObjB,X->Y) -> ( Right X -> Right Y );  -- Right functor action on morphisms
    default {
        unit  (A:ObjA):(           A -> Right Left A)    == { iLA(l:Left  A):Left  A == l; >>(      A,Left A,iLA)       }
        counit(B:ObjB):(Left Right B ->            B)    == { iRB(r:Right B):Right B == r; <<(Right B,     B,iRB)       }
        left (X:ObjA,Y:ObjA,f:X->Y):( Left X ->  Left Y) == { <<(      X,Left Y, (x:X):Right Left Y +-> unit(Y)(f x))   }
        right(X:ObjB,Y:ObjB,f:X->Y):(Right X -> Right Y) == { >>(Right X,     Y, (x:Left Right X):Y +-> f counit(X)(x)) }
    }
}

define LeftAdjoint(ObjA:Category,ObjB:Category,Right:ObjB->ObjA):Category == with {
    Left: ObjA -> ObjB;
    Adjoint(ObjA,ObjB,Left,Right)
}
define RightAdjoint(ObjA:Category,ObjB:Category,Left:ObjA->ObjB):Category == with {
    Right: ObjB -> ObjA;
    Adjoint(ObjA,ObjB,Left,Right)
}
define FreeConstruction(ObjA:Category,ObjB:Category,Forget:ObjB->ObjA):Category == with {
    Free: ObjA -> ObjB;
    Adjoint(ObjA,ObjB,Free,Forget)
}

define (Obj:Category)^(n:SingleInteger):Category == with {
    dom: Tuple Obj;
    put: dom -> %;
    get: % -> dom
}

nTuple(Obj:Category,n:SingleInteger)(t:Tuple Obj):(Obj^n) == add {
    dom:Tuple Obj == t;
    Rep == dom;
    put(d:dom):% == per d;
    get(d:%):dom == rep d
}
Diagonal(Obj:Category,n:SingleInteger)(A:Obj):(Obj^n) == nTuple(Obj,n)((A for i:SingleInteger in 1..n)@Tuple Obj) add;

import from SingleInteger;

define Equ(Obj:Category):Category == (Obj^2) with {
    import from Tuple Obj;
    one: element(dom,1) -> element(dom,2);
    two: element(dom,1) -> element(dom,2)
}

define CoEqu(Obj:Category):Category == (Obj^2) with {
    import from Tuple Obj;
    one: element(dom,2) -> element(dom,1);
    two: element(dom,2) -> element(dom,1)
}

define Pull(Obj:Category):Category == (Obj^3) with {
    import from Tuple Obj;
    one: element(dom,1) -> element(dom,2);
    two: element(dom,3) -> element(dom,2)
}

define Push(Obj:Category):Category == (Obj^3) with {
    import from Tuple Obj;
    one: element(dom,2) -> element(dom,1);
    two: element(dom,2) -> element(dom,3)
}

define   Product(Obj:Category):Category == with {  Product: (t:Tuple Obj) -> RightAdjoint(Obj, Obj^(length t), Diagonal(Obj,length t)) }
define CoProduct(Obj:Category):Category == with {CoProduct: (t:Tuple Obj) ->  LeftAdjoint(Obj^(length t), Obj, Diagonal(Obj,length t)) }
define     Final(Obj:Category):Category == with {    Final:            () -> RightAdjoint(Obj,Obj^0,Diagonal(Obj,0)) }
define   Initial(Obj:Category):Category == with {  Initial:            () ->  LeftAdjoint(Obj^0,Obj,Diagonal(Obj,0)) }
--define Equalizer(Obj:Category):Category == with {Equalizer: (A:Obj,B:Obj) -> RightAdjoint(Obj, Equ Obj, Diagonal(Obj,2)) }
       

